#!/usr/bin/env python3
#
# Written by Axel Hutt, INRIA Nancy Grand Est, France
# May 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


import matplotlib.pyplot as plt
import numpy as np
import matplotlib.mlab as mlab

class signal_class:
	def __init__(self):
		
		self.label = "timeseries_transient_oscillations.dat" # name of read data file
		
		# original data has sampling rate fs=30Hz
		self.fs = -1 # in Hz
		
		# original data has 3000 time points, i.e. duration T=100s
		self.T  = 100 # in seconds
		self.T_zeropadding = 00.0 
		
		self.dt_original          = -1.0 # dummy place holder for original data sampling time in seconds
		self.num_time_original    = -1 # dummy place holder for number of time points 
		
		## window flag
		## 0: rectangular window (no windowing)
		## 1: Hann window
		self.FLAG_WINDOW = 0
		
		## bandpass filter parameters
		## 0.4Hz band
		#self.f0   = 0.4 # filter frequency
		#self.B    = 0.2 # bandwidth
		
		## 1.0Hz band
		self.f0   = 1.0 # filter frequency
		self.B    = 0.5 # bandwidth

		## 1.0Hz band, filter diverges
		#self.f0   = 1.0 # filter frequency
		#self.B    = 0.25 # bandwidth
		
	
		self.w0   = 2.0*np.pi*self.f0
		self.Q    = self.w0/self.B # quality factor
		self.B2pi = self.B*2*np.pi
		self.xfilter = 0.0
		self.yfilter = 0.0
		
	def read_data(self):
		read_data=np.loadtxt(self.label)
		self.dt_original = read_data[1,0]-read_data[0,0]
		self.num_time_original = int(self.T/self.dt_original)
		self.data_original = read_data[:self.num_time_original,1]
		self.time_original = read_data[:self.num_time_original,0]
		print("original #steps:%d duration:%f"%(self.num_time_original,self.T))
	
	def window(self,i):
		zeropadding_num = int(self.T_zeropadding/(self.dt))
		if i<zeropadding_num//2 or i>=self.num_time-zeropadding_num//2:
			return 0.0
		else:
			if self.FLAG_WINDOW == 0:## rectangular
				return 1.0
			if self.FLAG_WINDOW == 1:## Hann
				return np.hanning(self.num_time-zeropadding_num)[i-zeropadding_num//2]
	
	def sample_signal(self):
		if self.fs < 0:
			self.fs = 1.0/self.dt_original
			sample_factor = 1
		self.dt       = 1.0/self.fs
		self.num_time = int(self.T/self.dt)
		print("original sample time: %f new sample time: %f   new #steps:%d"%(self.dt_original,self.dt,self.num_time))
		self.signal   = np.zeros(self.num_time)
		self.signal_f = np.zeros(self.num_time)
		self.time     = np.zeros(self.num_time)
		i = 0
		for ii in range(self.num_time_original):
			if ii%sample_factor == 0 and i<self.num_time:
				self.signal[i] = self.data_original[ii]
				self.time[i]   = self.time_original[ii]
				i += 1
		## apply zer0 padding and window function
		for i in range(self.num_time):
				self.signal[i]=self.signal[i]*self.window(i)
	
		
	def ode(self,x,y,I):
		xp = y
		yp = -self.B2pi*y-self.w0**2*x+self.B2pi*I
		self.xfilter = x + self.dt*xp
		self.yfilter = y + self.dt*yp
		
	def filter(self,I1,I0):
		dI = (I1-I0)/self.dt
		self.ode(self.xfilter,self.yfilter,dI)
		return self.xfilter
	
	def filter_signal(self):
		for i in range(self.num_time):
			self.signal_f[i] = self.filter(self.signal[i],self.signal[i-1])
	
	
	def compute_power(self):
		# frequency resolution of power spectral density
		df0 = 0.02       # useful for Bartlett PSD
		#df0 = 1.0/self.T # yields periodogram
		
		T=1.0/df0
		number_of_minimum_segments = 1
		num_perseg   = int(number_of_minimum_segments*T/self.dt) # number of data in single segment
		df = df0/float(number_of_minimum_segments)
		print("frequency resolution of PSD:%f"%df)
		
		overlap_percentage = 0.9 # 1.0: 100% overlap
		
		[power0,freqs0]=mlab.psd(\
				self.signal,\
				NFFT=num_perseg,\
				Fs=self.fs,\
				detrend=mlab.detrend_none,\
				window=mlab.window_hanning,\
				noverlap=num_perseg*overlap_percentage,\
				pad_to=None,\
				sides='onesided',\
				scale_by_freq=True)
		
		## frequency limitation of plotted PSD 
		fmin = 0.0
		fmax = 1.5		
		self.fmin_num=int(fmin/df)
		self.fmax_num=int(fmax/df)
		self.freqs=freqs0[self.fmin_num:self.fmax_num]
		self.power=power0[self.fmin_num:self.fmax_num]
	
	def plot(self):
		fign=plt.figure(1)
		
		ax = fign.add_subplot(211)
		plt.plot(self.time,self.signal,'k',self.time,self.signal_f,'r')
		plt.xlabel("time [s]")
		plt.ylabel("signal")
		
		ax = fign.add_subplot(212)
		plt.plot(self.freqs,self.power,'k')
		plt.xlabel("frequency [Hz]")
		plt.ylabel("|DFT|")
		
		fign.subplots_adjust(hspace=0.5)
		
		plt.show()
		
S = signal_class()
S.read_data()
S.sample_signal()
S.compute_power()
S.filter_signal()
S.plot()