#!/usr/bin/env python3
#
# Written by Axel Hutt, INRIA Nancy Grand Est, France
# May 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import matplotlib.pyplot as plt
from scipy.fft import fft, ifft, fftfreq
import numpy as np

class signal_class:
	def __init__(self):
		
		self.label = "FHN.dat" # name of read data file
		
		# original data has sampling rate fs=10Hz
		self.fs = 0.5 # in Hz
		
		# original data has 99999 time points, i.e. duration T=9999s
		self.T  = 500.0 # in seconds
		self.T_zeropadding = 000.0 
		
		self.dt_original          = -1.0 # dummy place holder for original data sampling time in seconds
		self.num_time_original    = -1 # dummy place holder for number of time points 
		
		
	def read_data(self):
		read_data=np.loadtxt(self.label)
		self.dt_original = read_data[1,0]-read_data[0,0]
		self.num_time_original = int(self.T/self.dt_original)
		self.data_original = read_data[:self.num_time_original,1]
		self.time_original = read_data[:self.num_time_original,0]
		print("original #steps:%d duration:%f"%(self.num_time_original,self.T))
		
	def sample_signal(self):
		self.dt       = 1.0/self.fs
		self.num_time = int(self.T/self.dt)
		sample_factor = self.dt/self.dt_original
		print("original sample time: %f new sample time: %f   new #steps:%d"%(self.dt_original,self.dt,self.num_time))
		self.signal = np.zeros(self.num_time)
		self.time   = np.zeros(self.num_time)
		i = 0
		for ii in range(self.num_time_original):
			if ii%sample_factor == 0 and i<self.num_time:
				self.signal[i] = self.data_original[ii]
				self.time[i]   = self.time_original[ii]
				i += 1
		## shorten
		zeropadding_num = int(self.T_zeropadding/(2.0*self.dt))
		for i in range(self.num_time):
			if i<zeropadding_num or i>=self.num_time-zeropadding_num:
				self.signal[i]=0.0
				
	def compute_DFT(self):
		DFT = fft(self.signal)
		self.DFT_of_signal = DFT[:self.num_time//2]
		self.freqs = fftfreq(self.num_time, self.dt)[:self.num_time//2]
		self.power = np.abs(self.DFT_of_signal)
	
	def write_out(self):
		
		label_ts = 'time_series.dat'
		label_ps = 'frequency_spectrum.dat'
		f_ts=open(label_ts,'w+')
		f_ps=open(label_ps,'w+')
		for i in range(self.num_time):
			str = '%f %f\n'%(self.time[i],self.signal[i])
			f_ts.write(str)
		for i in range(self.num_time//2):
			str = '%f %f\n'%(self.freqs[i],self.power[i])
			f_ps.write(str)
		f_ts.close()
		f_ps.close()

	def plot(self):
		fign=plt.figure(2)
		
		ax = fign.add_subplot(211)
		plt.plot(self.time,self.signal,'k')
		plt.xlabel("time [s]")
		plt.ylabel("signal")
		
		ax = fign.add_subplot(212)
		plt.plot(self.freqs,self.power,'k')
		plt.xlabel("frequency [Hz]")
		plt.ylabel("|DFT|")
		
		fign.subplots_adjust(hspace=0.5)
		plt.show()
		
S = signal_class()
S.read_data()
S.sample_signal()
S.compute_DFT()
S.write_out()
S.plot()