#!/usr/bin/env python3
#
# Written by Axel Hutt, INRIA Nancy Grand Est, France
# May 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


import matplotlib.pyplot as plt
import numpy as np
import matplotlib.mlab as mlab
from scipy.fft import fft, ifft
import scipy.fftpack as fftpack
import scipy.signal.windows as windows

class signal_class:
	def __init__(self):
		
		self.label = "timeseries_alternating_AlphaGamma.dat" # name of read data file
		
		# original data has sampling rate fs=2000Hz
		self.fs = 250 # in Hz
		
		# original data has 240000 time points, i.e. duration T=120s
		self.T  = 100 # in seconds
		self.T_zeropadding = 00.0 
		
		self.dt_original          = -1.0 # dummy place holder for original data sampling time in seconds
		self.num_time_original    = -1 # dummy place holder for number of time points 
		
		## window flag
		## 0: rectangular window (no windowing)
		## 1: Hann window
		self.FLAG_WINDOW = 0
		
		
	def read_data(self):
		read_data=np.loadtxt(self.label)
		self.dt_original = read_data[1,0]-read_data[0,0]
		self.num_time_original = int(self.T/self.dt_original)
		self.data_original = read_data[:self.num_time_original,1]
		self.time_original = read_data[:self.num_time_original,0]
		print("original #steps:%d duration:%f"%(self.num_time_original,self.T))
	
	def window(self,i):
		zeropadding_num = int(self.T_zeropadding/(self.dt))
		if i<zeropadding_num//2 or i>=self.num_time-zeropadding_num//2:
			return 0.0
		else:
			if self.FLAG_WINDOW == 0:## rectangular
				return 1.0
			if self.FLAG_WINDOW == 1:## Hann
				return np.hanning(self.num_time-zeropadding_num)[i-zeropadding_num//2]
	
	def sample_signal(self):
		if self.fs < 0:
			self.fs = 1.0/self.dt_original
			sample_factor = 1
		else:
			sample_factor = 1.0/(self.dt_original*self.fs)
		
		self.dt       = 1.0/self.fs
		self.num_time = int(self.T/self.dt)
		print("original sample time: %f new sample time: %f   new #steps:%d"%(self.dt_original,self.dt,self.num_time))
		self.signal   = np.zeros(self.num_time)
		self.time     = np.zeros(self.num_time)
		i = 0
		for ii in range(self.num_time_original):
			if ii%sample_factor == 0 and i<self.num_time:
				self.signal[i] = self.data_original[ii]
				self.time[i]   = self.time_original[ii]
				i += 1
		## apply zer0 padding and window function
		for i in range(self.num_time):
				self.signal[i]=self.signal[i]*self.window(i)
	
	def STFT(self):
		
		window_duration		= 2.0 # window lengths in seconds
		
		# number of shifts of window (in time points)
		#num_shift			= 200 # corresponds to 13.4s overlap
		num_shift			= 10 # corresponds to 19.7s overlap
		
		num_window			= int(window_duration/self.dt)
		num_times_nooverlap	= int(self.num_time/num_window)
		num_perwindow		= int(num_window/num_shift)
		num_times			= num_perwindow*(num_times_nooverlap-1)# how many time steps of duration num_shift*dt
		
		fmin_move = 5.0 # lower border of visualized frequency
		fmax_move = 50.0 # upper border of visualized frequency
		df_move   = 1.0/window_duration
		fmin_num  = int(fmin_move/df_move)
		fmax_num  = int(fmax_move/df_move)
		num_freq  = fmax_num-fmin_num
		
		Xfft_move    = np.zeros((num_times,num_window))
		Xabsfft_move = np.zeros((num_times,num_window))
		j0 			 = num_window//2
		for j in range(0,num_times):
			jstart 			  = j0+j*num_shift-num_window//2
			jend   			  = j0+j*num_shift+num_window//2
			#sig    			  = self.signal[jstart:jend]*np.hanning(jend-jstart) # Hanning window
			sig    			  = self.signal[jstart:jend]*windows.gaussian(num_window,num_window//4) # Gaussian window
			Xabsfft_move[j,:] = np.abs(fftpack.fft(sig,num_window))/window_duration
			
		self.initialtime=self.dt*j0
		self.endtime = (j0+(num_times-1)*num_shift)*self.dt
		self.power_ts  = np.transpose(Xabsfft_move[:,fmin_num:fmax_num])
		self.power_ts_log = np.log(self.power_ts)
		self.fmin  = fmin_move
		self.fmax  = fmax_move
		
	
	def Wavelettransform_Morlet(self):
		
		self.sigma = 8.0
		center_frequency = self.sigma/(2.0*np.pi)
		self.df    = 1.0/self.T
		
		def FT_nonadmissible_motherwavelet(k):
			nu = k*self.df
			return np.sqrt(np.pi)*np.exp(-0.25*(self.sigma-2.0*np.pi*nu)**2)
		
		def FT_complexmotherwavelet(k):
			nu = k*self.df
			return np.sqrt(np.pi)*(np.exp(-0.25*(self.sigma-2.0*np.pi*nu)**2)-np.exp(-0.25*self.sigma**2)*np.exp(-np.pi**2*nu**2))
		
		def FT_motherwavelet(k):
			#return FT_nonadmissible_motherwavelet(k)
			return FT_complexmotherwavelet(k)
	
		self.fmin = 5.0  # minimum frequency
		self.fmax = 50.0  # maximum frequency
		self.num_scales = 20 # number of scales
		freqs = np.linspace(self.fmin,self.fmax,self.num_scales)
		self.a = center_frequency/(freqs)
		
		DFT_convolved = np.zeros(self.num_time,dtype=complex)
		self.WT = np.zeros((self.num_scales,self.num_time))
		
		DFT = fft(self.signal)
		
		for l in range(self.num_scales):
			for k in range(self.num_time//2):
				DFT_convolved[k] = DFT[k]*FT_motherwavelet(self.a[l]*k)
				kk = k+self.num_time//2
				DFT_convolved[kk] = DFT[kk]*FT_motherwavelet(self.a[l]*(-self.num_time//2+k))
			self.WT[l,:] = np.abs(ifft(DFT_convolved))
			
		self.initialtime=0.0
		self.endtime = self.num_time*self.dt
		
	def plot(self):
		
		def forceAspect(ax,aspect=1):
			im = ax.get_images()
			extent =  im[0].get_extent()
			ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)
		
		fign=plt.figure(1)
		
		ax = fign.add_subplot(221)
		plt.plot(self.time,self.signal,'k')
		plt.xlabel("time [s]")
		plt.ylabel("signal")
		
		
		ax = fign.add_subplot(223)
		plt.imshow(self.WT, interpolation='nearest', cmap=plt.cm.jet,origin='lower',extent=[self.initialtime,self.endtime, self.fmin, self.fmax])		
		forceAspect(ax,aspect=1)
		plt.colorbar()
		plt.xlabel("time [s]")
		plt.ylabel("frequency [Hz]")
		
		ax = fign.add_subplot(224)
		plt.imshow(self.power_ts_log, interpolation='nearest', cmap=plt.cm.jet,origin='lower',extent=[self.initialtime,self.endtime, self.fmin, self.fmax])	
		forceAspect(ax,aspect=1)
		plt.colorbar()
		plt.xlabel("time [s]")
		plt.ylabel("frequency [Hz]")
		
		
		fign.subplots_adjust(hspace=0.5)
		
		plt.show()
		
S = signal_class()
S.read_data()
S.sample_signal()
S.Wavelettransform_Morlet()
S.STFT()
S.plot()