#!/usr/bin/env python3
#
# Written by Axel Hutt, INRIA Nancy Grand Est, France
# May 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


import matplotlib.pyplot as plt
import numpy as np
import matplotlib.mlab as mlab
from scipy.fft import fft, ifft
import emd

class signal_class:
	def __init__(self):
		
		self.fs = 200 # in Hz
		self.dt = 1.0/self.fs
		self.T  = 1.0 # in seconds
		self.num_time = int(self.T*self.fs)
		
	def create_signal(self):
		self.time 	= np.linspace(0, self.T, self.num_time)
		self.signal = np.cos(11*2*np.pi*self.time**2) + 6*self.time**2
	
	def Wavelettransform_Morlet(self):
		
		self.sigma = 8.0
		center_frequency = self.sigma/(2.0*np.pi)
		self.df    = 1.0/self.T
		
		def FT_complexmotherwavelet(k):
			nu = k*self.df
			return np.sqrt(np.pi)*(np.exp(-0.25*(self.sigma-2.0*np.pi*nu)**2)-np.exp(-0.25*self.sigma**2)*np.exp(-np.pi**2*nu**2))
		
		def FT_motherwavelet(k):
			return FT_complexmotherwavelet(k)
		
		self.fmin 		= 5.0  # minimum frequency
		self.fmax 		= 25.0  # maximum frequency
		self.num_scales = 20 # number of scales
		freqs 			= np.linspace(self.fmin,self.fmax,self.num_scales)
		self.a 			= center_frequency/(freqs)
		
		DFT_convolved = np.zeros(self.num_time,dtype=complex)
		self.WT = np.zeros((self.num_scales,self.num_time))
		
		DFT = fft(self.signal)
		
		for l in range(self.num_scales):
			for k in range(self.num_time//2):
				DFT_convolved[k] = DFT[k]*FT_motherwavelet(self.a[l]*k)
				kk = k+self.num_time//2
				DFT_convolved[kk] = DFT[kk]*FT_motherwavelet(self.a[l]*(-self.num_time//2+k))
			self.WT[l,:] = np.abs(ifft(DFT_convolved))
			
		self.initialtime=0.0
		self.endtime = self.num_time*self.dt
	
	def EMD(self):
		self.IMFS = emd.sift.sift(self.signal)
		self.N = np.shape(self.IMFS)[1]
		self.IP, self.IF, self.IA = emd.spectra.frequency_transform(self.IMFS, self.fs, 'nht')
		self.freq_edges, self.freq_bins = emd.spectra.define_hist_bins(0, 10, 50)
		self.hht = emd.spectra.hilberthuang(self.IF, self.IA, self.freq_edges)
		
	def plot(self):
		
		def forceAspect(ax,aspect=1):
			im = ax.get_images()
			extent =  im[0].get_extent()
			ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)
		
		
		fign=plt.figure(1)
		plt.subplot(self.N+1,3,1)
		plt.plot(self.time, self.signal, 'r')
		plt.title("signal")
		plt.xlabel("Time [s]")
		plt.ylabel("signal")
		
		ax = plt.subplot(self.N+1,3,2)
		plt.title("Morlet wavelet transform")
		plt.imshow(self.WT, interpolation='nearest', cmap=plt.cm.jet,origin='lower',extent=[self.initialtime,self.endtime, self.fmin, self.fmax])
		plt.xlabel("Time [s]")
		plt.ylabel("Freq [Hz]")
		forceAspect(ax,aspect=2)
		
		
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+4) 
			plt.plot(self.time, self.IMFS[:,n], 'g') 
			plt.title("IMF "+str(n+1)) 
			plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+5) 
			plt.plot(self.time, self.IF[:,n], 'k') 
			plt.title("IF"+str(n+1)) 
			plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+6) 
			plt.plot(self.time, self.IA[:,n], 'g') 
			plt.title("IA "+str(n+1)) 
			plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		fign.tight_layout()
		plt.show()
		
S = signal_class()
S.create_signal()
S.Wavelettransform_Morlet()
S.EMD()
S.plot()