#!/usr/bin/env python3
#
# Written by Axel Hutt, INRIA Nancy Grand Est, France
# May 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


import matplotlib.pyplot as plt
import numpy as np
import matplotlib.mlab as mlab
from scipy.fft import fft, ifft
import emd

class signal_class:
	def __init__(self):
		
		self.label = "timeseries_transient_oscillations.dat" # name of read data file
		
		# original data has sampling rate fs=30Hz
		self.fs = -1 # in Hz
		
		# original data has 3000 time points, i.e. duration T=100s
		self.T  = 100 # in seconds
		self.T_zeropadding = 00.0 
		
		self.dt_original          = -1.0 # dummy place holder for original data sampling time in seconds
		self.num_time_original    = -1 # dummy place holder for number of time points 
		
		## window flag
		## 0: rectangular window (no windowing)
		## 1: Hann window
		self.FLAG_WINDOW = 0
		
		
	def read_data(self):
		read_data=np.loadtxt(self.label)
		self.dt_original = read_data[1,0]-read_data[0,0]
		self.num_time_original = int(self.T/self.dt_original)
		self.data_original = read_data[:self.num_time_original,1]
		self.time_original = read_data[:self.num_time_original,0]
		print("original #steps:%d duration:%f"%(self.num_time_original,self.T))
	
	def window(self,i):
		zeropadding_num = int(self.T_zeropadding/(self.dt))
		if i<zeropadding_num//2 or i>=self.num_time-zeropadding_num//2:
			return 0.0
		else:
			if self.FLAG_WINDOW == 0:## rectangular
				return 1.0
			if self.FLAG_WINDOW == 1:## Hann
				return np.hanning(self.num_time-zeropadding_num)[i-zeropadding_num//2]
	
	def sample_signal(self):
		if self.fs < 0:
			self.fs = 1.0/self.dt_original
			sample_factor = 1
		self.dt       = 1.0/self.fs
		self.num_time = int(self.T/self.dt)
		print("original sample time: %f new sample time: %f   new #steps:%d"%(self.dt_original,self.dt,self.num_time))
		self.signal   = np.zeros(self.num_time)
		self.time     = np.zeros(self.num_time)
		i = 0
		for ii in range(self.num_time_original):
			if ii%sample_factor == 0 and i<self.num_time:
				self.signal[i] = self.data_original[ii]
				self.time[i]   = self.time_original[ii]
				i += 1
		## apply zer0 padding and window function
		for i in range(self.num_time):
				self.signal[i]=self.signal[i]*self.window(i)
	
	
	def Wavelettransform_Morlet(self):
		
		self.sigma = 10.0
		center_frequency = self.sigma/(2.0*np.pi)
		self.df    = 1.0/self.T
		
		def FT_realmotherwavelet(k):
			nu = k*self.df
			return np.sqrt(np.pi)*np.exp(-0.25*(self.sigma-2.0*np.pi*nu)**2)
		
		def FT_complexmotherwavelet(k):
			nu = k*self.df
			return np.sqrt(np.pi)*(np.exp(-0.25*(self.sigma-2.0*np.pi*nu)**2)-np.exp(-0.25*self.sigma**2)*np.exp(-np.pi**2*nu**2))
		
		def FT_motherwavelet(k):
			return FT_realmotherwavelet(k)
			#return FT_complexmotherwavelet(k)
	
		self.fmin = 0.2  # minimum frequency
		self.fmax = 1.3  # maximum frequency
		self.num_scales = 20 # number of scales
		freqs = np.linspace(self.fmin,self.fmax,self.num_scales)
		self.a = center_frequency/(freqs)
		DFT_convolved = np.zeros(self.num_time,dtype=complex)
		self.WT = np.zeros((self.num_scales,self.num_time))
		
		DFT = fft(self.signal)
		
		for l in range(self.num_scales):
			for k in range(self.num_time//2):
				DFT_convolved[k] = DFT[k]*FT_motherwavelet(self.a[l]*k)
				kk = k+self.num_time//2
				DFT_convolved[kk] = DFT[kk]*FT_motherwavelet(self.a[l]*(-self.num_time//2+k))
			self.WT[l,:] = np.abs(ifft(DFT_convolved))
			
		self.initialtime=0.0
		self.endtime = self.num_time*self.dt
		
	def EMD(self):
		self.IMFS = emd.sift.sift(self.signal)
		self.N = np.shape(self.IMFS)[1]
		self.IP, self.IF, self.IA = emd.spectra.frequency_transform(self.IMFS, self.fs, 'nht')
		self.freq_edges, self.freq_bins = emd.spectra.define_hist_bins(0, 10, 50)
		self.hht = emd.spectra.hilberthuang(self.IF, self.IA, self.freq_edges)
		
	def plot(self):
		
		def forceAspect(ax,aspect=1):
			im = ax.get_images()
			extent =  im[0].get_extent()
			ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)
			
			
		fign=plt.figure(1)
		plt.subplot(self.N+1,3,1)
		plt.plot(self.time, self.signal, 'r')
		#plt.title("signal")
		#plt.xlabel("Time [s]")
		#plt.ylabel("signal")
		
		ax = plt.subplot(self.N+1,3,2)
		#plt.title("Morlet wavelet transform")
		plt.imshow(self.WT, interpolation='nearest', cmap=plt.cm.jet,origin='lower',extent=[self.initialtime,self.endtime, self.fmin, self.fmax])
		#plt.xlabel("Time [s]")
		#plt.ylabel("Freq [Hz]")
		forceAspect(ax,aspect=8)
		
		
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+4) 
			plt.plot(self.time, self.IMFS[:,n], 'g') 
			#plt.title("IMF "+str(n+1)) 
			#plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+5) 
			plt.plot(self.time, self.IF[:,n], 'k') 
			#plt.title("IF"+str(n+1)) 
			#plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		for n in range(self.N): 
			plt.subplot(self.N+1,3,3*n+6) 
			plt.plot(self.time, self.IA[:,n], 'g') 
			#plt.title("IA "+str(n+1)) 
			#plt.xlabel("Time [s]")
			plt.xlim(self.time[0], self.time[-1])
			plt.grid(True)
		#fign.tight_layout()
		fign.subplots_adjust(hspace=0.5)
		
		
		plt.show()
		
	
		
S = signal_class()
S.read_data()
S.sample_signal()
S.Wavelettransform_Morlet()
S.EMD()
S.plot()